import numpy as np


class OneHotEncoder:
    def __init__(self):
        self.data = None
        self.n_classes = 0
        self.labels = None
        self.label_to_vec = {}
        self.int_to_label = {}

    def fit(self, data):
        data = np.array(data).flatten()
        self.labels = set(data)
        self.n_classes = len(self.labels)
        for index, label in enumerate(self.labels):
            vec = np.array([0] * self.n_classes)
            vec[index] = 1
            self.label_to_vec[label] = vec
            self.int_to_label[index] = label

    def transform(self, data):
        data = np.array(data)
        return np.array(list(map(lambda x: self.label_to_vec[x], data)))

    def inverse_transform(self, data):
        return np.array(list(map(lambda x: self.int_to_label[x], np.argmax(np.array(data), axis=1))))



import numpy as np
import pickle
import os

from sklearn.metrics import accuracy_score
from sklearn.model_selection import train_test_split, StratifiedKFold

from autokeras import constant
from autokeras.search import HillClimbingSearcher, RandomSearcher
from autokeras.preprocessor import OneHotEncoder
from autokeras.utils import ensure_dir, reset_weights, ModelTrainer


def load_from_path(path=constant.DEFAULT_SAVE_PATH):
    classifier = pickle.load(open(os.path.join(path, 'classifier'), 'rb'))
    classifier.path = path
    classifier.searcher = pickle.load(open(os.path.join(path, 'searcher'), 'rb'))
    return classifier


class ClassifierBase:
    def __init__(self, verbose=False, searcher_type=None, path=constant.DEFAULT_SAVE_PATH):
        self.y_encoder = None
        self.verbose = verbose
        self.searcher = None
        self.searcher_type = searcher_type
        # self.history = []
        self.path = path
        self.model_id = None
        ensure_dir(path)

    def _validate(self, x_train, y_train):
        try:
            x_train = x_train.astype('float64')
        except ValueError:
            raise ValueError('x_train should only contain numerical data.')

        if len(x_train.shape) < 2:
            raise ValueError('x_train should at least has 2 dimensions.')

        if x_train.shape[0] != y_train.shape[0]:
            raise ValueError('x_train and y_train should have the same number of instances.')

    def fit(self, x_train, y_train):
        x_train = np.array(x_train)
        y_train = np.array(y_train).flatten()

        self._validate(x_train, y_train)

        # Transform y_train.
        if self.y_encoder is None:
            self.y_encoder = OneHotEncoder()
            self.y_encoder.fit(y_train)

        y_train = self.y_encoder.transform(y_train)

        if self.searcher is None:
            input_shape = x_train.shape[1:]
            n_classes = self.y_encoder.n_classes
            self.searcher = self._get_searcher_class()(n_classes, input_shape, self.path, self.verbose)

        # Divide training data into training and testing data.
        x_train, x_test, y_train, y_test = train_test_split(x_train, y_train, test_size=0.33, random_state=42)

        pickle.dump(self, open(os.path.join(self.path, 'classifier'), 'wb'))
        self.model_id = self.searcher.search(x_train, y_train, x_test, y_test)

    def predict(self, x_test):
        model = self.searcher.load_best_model()
        return self.y_encoder.inverse_transform(model.predict(x_test, verbose=self.verbose))

    def summary(self):
        model = self.searcher.load_best_model()
        model.summary()

    def _get_searcher_class(self):
        if self.searcher_type == 'climb':
            return HillClimbingSearcher
        elif self.searcher_type == 'random':
            return RandomSearcher
        return None

    def evaluate(self, x_test, y_test):
        y_predict = self.predict(x_test)
        return accuracy_score(y_test, y_predict)

    def cross_validate(self, x_all, y_all, n_splits):
        k_fold = StratifiedKFold(n_splits=n_splits, shuffle=False, random_state=7)
        scores = []
        y_raw_all = y_all
        y_all = self.y_encoder.transform(y_all)
        for train, test in k_fold.split(x_all, y_raw_all):
            model = self.searcher.load_best_model()
            reset_weights(model)
            ModelTrainer(model, x_all[train], y_all[train], x_all[test], y_all[test], self.verbose).train_model()
            scores = model.evaluate(x_all[test], y_all[test], verbose=self.verbose)
            scores.append(scores[1] * 100)
        return np.array(scores)


class Classifier(ClassifierBase):
    def __init__(self):
        super().__init__()

    def _validate(self, x_train, y_train):
        super()._validate(x_train, y_train)


class ImageClassifier(ClassifierBase):
    def __init__(self, verbose=True, searcher_type='climb', path=constant.DEFAULT_SAVE_PATH):
        super().__init__(verbose, searcher_type, path)



import numpy as np
from keras.layers import Dense
from keras.models import Sequential

from autokeras.constant import CONV_FUNC_LIST, WEIGHTED_LAYER_FUNC_LIST
from autokeras.net_transformer import to_deeper_model
from autokeras.utils import is_conv_layer, copy_layer


def layer_num(model, type_list):
    return len(list(filter(lambda layer: isinstance(layer, tuple(type_list)), model.layers)))


def add_dense(net, n_add):
    level = None
    for index, layer in enumerate(net.layers):
        if isinstance(layer, Dense):
            level = index
            break
    for i in range(n_add):
        net = to_deeper_model(net, level)
    return net


def add_conv(net, n_add):
    level = None
    for index, layer in enumerate(net.layers):
        if is_conv_layer(layer):
            level = index
            break
    for i in range(n_add):
        net = to_deeper_model(net, level)
    return net


def pad_filter(weight, old_size, new_size):
    pad_width = tuple(map(lambda x: (int(x), int(x)), np.subtract(np.array(new_size), np.array(old_size)) / 2))
    pad_width += ((0, 0), (0, 0))
    return np.pad(weight,
                  pad_width,
                  'constant',
                  constant_values=0)


def combine_conv_weights(layer1, layer2, filter_size, n_input_channel):
    weight1, bias1 = layer1.get_weights()
    weight2, bias2 = layer2.get_weights()

    n_filters1 = layer1.filters
    n_filters2 = layer2.filters

    old_input_channel1 = weight1.shape[-2]
    old_input_channel2 = weight2.shape[-2]

    pad_weight1 = pad_filter(weight1, layer1.kernel_size, filter_size)
    pad_weight1 = np.concatenate((pad_weight1,
                                  np.zeros(filter_size + (n_input_channel - old_input_channel1, n_filters1))),
                                 axis=-2)
    pad_weight2 = pad_filter(weight2, layer2.kernel_size, filter_size)
    pad_weight2 = np.concatenate((np.zeros(filter_size + (n_input_channel - old_input_channel2, n_filters2)),
                                  pad_weight2),
                                 axis=-2)
    return np.concatenate((pad_weight1, pad_weight2), axis=-1), np.concatenate((bias1, bias2))


def combine_last_layer(layer1, layer2):
    pre_units1 = layer1.get_weights()[0].shape[0]
    pre_units2 = layer2.get_weights()[0].shape[0]

    weight1, bias1 = layer1.get_weights()
    weight2, bias2 = layer2.get_weights()

    new_layer = copy_layer(layer1)
    new_layer.build((None, pre_units1 + pre_units2))
    new_weights = np.concatenate((weight1, weight2), axis=0)
    new_bias = np.mean((bias1, bias2), axis=0)
    new_layer.set_weights((new_weights, new_bias))

    return new_layer


class NetCombinator:
    def __init__(self):
        self.input_shape = None

    def get_output_channel(self, layer_list):
        last_conv = None
        for layer in layer_list:
            if is_conv_layer(layer):
                last_conv = layer

        if not last_conv:
            return self.input_shape[-1]

        return last_conv.filters

    def combine_layer(self, layer_list, layer1, layer2):
        if is_conv_layer(layer1):
            filter1 = layer1.kernel_size
            filter2 = layer2.kernel_size
            new_filter = tuple(np.array([filter1, filter2]).max(axis=1))
            layer = layer1.__class__(layer1.filters + layer2.filters,
                                     kernel_size=new_filter,
                                     padding='same',
                                     activation='relu')
            n_input_channel = self.get_output_channel(layer_list)
            layer.build((None,) * len(new_filter) + (n_input_channel,))
            layer.set_weights = combine_conv_weights(layer1, layer2, new_filter, n_input_channel)
            return layer

        if isinstance(layer1, Dense):
            weight1, bias1 = layer1.get_weights()
            weight2, bias2 = layer2.get_weights()

            unit1 = layer1.units
            unit2 = layer2.units

            pre_unit1 = layer1.get_weights()[0].shape[0]
            pre_unit2 = layer2.get_weights()[0].shape[0]

            new_layer = Dense(unit1 + unit2, activation='relu')
            new_layer.build((None, pre_unit1 + pre_unit2))
            new_weight = np.zeros((pre_unit1 + pre_unit2, unit1 + unit2))
            new_weight[:pre_unit1, :unit1] = weight1
            new_weight[pre_unit1:, unit1:] = weight2
            new_bias = np.concatenate((bias1, bias2))

            new_layer.set_weights((new_weight, new_bias))
            return new_layer

        return copy_layer(layer1)

    def combine(self, net1, net2):
        self.input_shape = net1.layers[0].input_shape
        n_conv1 = layer_num(net1, CONV_FUNC_LIST)
        n_conv2 = layer_num(net2, CONV_FUNC_LIST)
        n_dense1 = layer_num(net1, [Dense])
        n_dense2 = layer_num(net2, [Dense])

        # Transform the two networks into same number of weighted layers.
        if n_conv1 > n_conv2:
            net2 = add_conv(net2, n_conv1 - n_conv2)
        elif n_conv2 > n_conv1:
            net1 = add_conv(net1, n_conv2 - n_conv1)

        if n_dense1 > n_dense2:
            net2 = add_dense(net2, n_dense1 - n_dense2)
        elif n_dense2 > n_dense1:
            net1 = add_dense(net2, n_dense2 - n_dense1)

        # Combine the corresponding layers.
        p1 = 0
        p2 = 0
        layers1 = net1.layers
        layers2 = net2.layers
        new_layer_list = []
        while p1 < len(layers1) - 1 and p2 < len(layers2) - 1:
            layer1 = layers1[p1]
            layer2 = layers2[p2]
            if type(layer1) == type(layer2):
                new_layer_list.append(self.combine_layer(new_layer_list, layer1, layer2))
                p1 += 1
                p2 += 1
            elif isinstance(layer1, tuple(WEIGHTED_LAYER_FUNC_LIST)):
                new_layer_list.append(copy_layer(layer2))
                p2 += 1
            elif isinstance(layer2, tuple(WEIGHTED_LAYER_FUNC_LIST)):
                new_layer_list.append(copy_layer(layer1))
                p1 += 1
            else:
                print(layer1, layer2)

        # Add the first layer with input_shape.
        new_model = Sequential()
        first_layer_config = new_layer_list[0].get_config()
        first_layer_config['input_shape'] = self.input_shape[1:]
        new_first_layer = new_layer_list[0].__class__.from_config(first_layer_config)
        new_model.add(new_first_layer)
        new_first_layer.set_weights(new_layer_list[0].get_weights())

        for layer in new_layer_list[1:]:
            new_model.add(layer)

        new_model.add(combine_last_layer(layers1[-1], layers2[-1]))

        return new_model


def combine(net1, net2):
    return NetCombinator().combine(net1, net2)



from functools import reduce
from operator import mul

import numpy as np
from keras.layers import Dense, Conv1D, Conv2D, Conv3D

from autokeras.utils import get_conv_layer_func


def conv_to_deeper_layer(conv_layer):
    filter_shape = conv_layer.kernel_size
    n_filters = conv_layer.filters
    weight = np.zeros(filter_shape + (n_filters, n_filters))
    center = tuple(map(lambda x: int((x - 1) / 2), filter_shape))
    for i in range(n_filters):
        filter_weight = np.zeros(filter_shape + (n_filters,))
        index = center + (i,)
        filter_weight[index] = 1
        weight[..., i] = filter_weight
    bias = np.zeros(n_filters)
    conv_func = get_conv_layer_func(len(filter_shape))
    new_conv_layer = conv_func(n_filters,
                               kernel_size=filter_shape,
                               activation='relu',
                               padding='same')
    new_conv_layer.build((None,) * (len(filter_shape) + 1) + (n_filters,))
    new_conv_layer.set_weights((weight, bias))
    return new_conv_layer


def dense_to_deeper_layer(dense_layer):
    units = dense_layer.units
    weight = np.eye(units)
    bias = np.zeros(units)
    new_dense_layer = Dense(units, activation='relu')
    new_dense_layer.build((None, units))
    new_dense_layer.set_weights((weight, bias))
    return new_dense_layer


def to_deeper_layer(layer):
    if isinstance(layer, Dense):
        return dense_to_deeper_layer(layer)
    elif isinstance(layer, (Conv1D, Conv2D, Conv3D)):
        return conv_to_deeper_layer(layer)
    raise ValueError("Layer must be a Dense or Convolution Layer")


def dense_to_wider_layer(pre_layer, next_layer, n_add_units):
    n_units1 = pre_layer.get_weights()[0].shape[0]
    n_units2 = pre_layer.units
    n_units3 = next_layer.units

    teacher_w1 = pre_layer.get_weights()[0]
    teacher_b1 = pre_layer.get_weights()[1]
    teacher_w2 = next_layer.get_weights()[0]
    teacher_b2 = next_layer.get_weights()[1]
    rand = np.random.randint(n_units2, size=n_add_units)
    replication_factor = np.bincount(rand)
    student_w1 = teacher_w1.copy()
    student_w2 = teacher_w2.copy()
    student_b1 = teacher_b1.copy()

    # target layer update (i)
    for i in range(n_add_units):
        teacher_index = rand[i]
        new_weight = teacher_w1[:, teacher_index]
        new_weight = new_weight[:, np.newaxis]
        student_w1 = np.concatenate((student_w1, new_weight), axis=1)
        student_b1 = np.append(student_b1, teacher_b1[teacher_index])

    # next layer update (i+1)
    for i in range(n_add_units):
        teacher_index = rand[i]
        n_copies = replication_factor[teacher_index] + 1
        new_weight = teacher_w2[teacher_index, :]*(1./n_copies)
        new_weight = new_weight[np.newaxis, :]
        student_w2 = np.concatenate((student_w2, new_weight), axis=0)
        student_w2[teacher_index, :] = new_weight

    new_pre_layer = Dense(n_units2 + n_add_units, input_shape=(n_units1,), activation='relu')
    new_pre_layer.build((None, n_units1))
    new_pre_layer.set_weights((student_w1, student_b1))
    new_next_layer = Dense(n_units3, activation=next_layer.get_config()['activation'])
    new_next_layer.build((None, n_units2 + n_add_units))
    new_next_layer.set_weights((student_w2, teacher_b2))

    return new_pre_layer, new_next_layer


def conv_to_wider_layer(pre_layer, next_layer, n_add_filters):
    pre_filter_shape = pre_layer.kernel_size
    next_filter_shape = next_layer.kernel_size
    conv_func = get_conv_layer_func(len(pre_filter_shape))
    n_pre_filters = pre_layer.filters
    n_next_filters = next_layer.filters

    teacher_w1 = pre_layer.get_weights()[0]
    teacher_w2 = next_layer.get_weights()[0]
    teacher_b1 = pre_layer.get_weights()[1]
    teacher_b2 = next_layer.get_weights()[1]

    rand = np.random.randint(n_pre_filters, size=n_add_filters)
    replication_factor = np.bincount(rand)
    student_w1 = teacher_w1.copy()
    student_w2 = teacher_w2.copy()
    student_b1 = teacher_b1.copy()
    # target layer update (i)
    for i in range(len(rand)):
        teacher_index = rand[i]
        new_weight = teacher_w1[..., teacher_index]
        new_weight = new_weight[..., np.newaxis]
        student_w1 = np.concatenate((student_w1, new_weight), axis=-1)
        student_b1 = np.append(student_b1, teacher_b1[teacher_index])
    # next layer update (i+1)
    for i in range(len(rand)):
        teacher_index = rand[i]
        factor = replication_factor[teacher_index] + 1
        new_weight = teacher_w2[..., teacher_index, :] * (1. / factor)
        new_weight_re = new_weight[..., np.newaxis, :]
        student_w2 = np.concatenate((student_w2, new_weight_re), axis=-2)
        student_w2[..., teacher_index, :] = new_weight

    print(pre_layer.input_shape)
    new_pre_layer = conv_func(n_pre_filters + n_add_filters,
                              kernel_size=pre_filter_shape,
                              activation='relu',
                              padding='same',
                              input_shape=pre_layer.input_shape[1:])
    new_pre_layer.build((None,) * (len(pre_filter_shape) + 1) + (pre_layer.input_shape[-1],))
    new_pre_layer.set_weights((student_w1, student_b1))
    new_next_layer = conv_func(n_next_filters, kernel_size=next_filter_shape, activation='relu', padding='same')
    new_next_layer.build((None,) * (len(next_filter_shape) + 1) + (n_pre_filters + n_add_filters,))
    new_next_layer.set_weights((student_w2, teacher_b2))
    return new_pre_layer, new_next_layer


def conv_dense_to_wider_layer(pre_layer, next_layer, n_add_filters):
    pre_filter_shape = pre_layer.kernel_size
    conv_func = get_conv_layer_func(len(pre_filter_shape))
    n_pre_filters = pre_layer.filters
    n_units = next_layer.units

    teacher_w1 = pre_layer.get_weights()[0]
    teacher_w2 = next_layer.get_weights()[0]
    n_total_weights = int(reduce(mul, teacher_w2.shape))
    teacher_w2 = teacher_w2.reshape(int(n_total_weights / n_pre_filters / n_units), n_pre_filters, n_units)
    teacher_b1 = pre_layer.get_weights()[1]
    teacher_b2 = next_layer.get_weights()[1]
    rand = np.random.randint(n_pre_filters, size=n_add_filters)
    replication_factor = np.bincount(rand)
    student_w1 = teacher_w1.copy()
    student_w2 = teacher_w2.copy()
    student_b1 = teacher_b1.copy()
    # target layer update (i)
    for i in range(len(rand)):
        teacher_index = rand[i]
        new_weight = teacher_w1[..., teacher_index]
        new_weight = new_weight[..., np.newaxis]
        student_w1 = np.concatenate((student_w1, new_weight), axis=-1)
        student_b1 = np.append(student_b1, teacher_b1[teacher_index])
    # next layer update (i+1)
    for i in range(len(rand)):
        teacher_index = rand[i]
        factor = replication_factor[teacher_index] + 1
        new_weight = teacher_w2[:, teacher_index, :] * (1. / factor)
        new_weight_re = new_weight[:, np.newaxis, :]
        student_w2 = np.concatenate((student_w2, new_weight_re), axis=1)
        student_w2[:, teacher_index, :] = new_weight

    new_pre_layer = conv_func(n_pre_filters + n_add_filters,
                              kernel_size=pre_filter_shape,
                              activation='relu',
                              padding='same',
                              input_shape=pre_layer.input_shape[1:])
    new_pre_layer.build((None,) * (len(pre_filter_shape) + 1) + (pre_layer.input_shape[-1],))
    new_pre_layer.set_weights((student_w1, student_b1))
    new_next_layer = Dense(n_units, activation=next_layer.get_config()['activation'])
    n_new_total_weights = int(reduce(mul, student_w2.shape))
    input_dim = int(n_new_total_weights / n_units)
    new_next_layer.build((None, input_dim))
    new_next_layer.set_weights((student_w2.reshape(input_dim, n_units), teacher_b2))
    return new_pre_layer, new_next_layer


def to_wider_layer(pre_layer, next_layer, n_add):
    if isinstance(pre_layer, (Conv1D, Conv2D, Conv3D)) and isinstance(next_layer, Dense):
        return conv_dense_to_wider_layer(pre_layer, next_layer, n_add)
    elif isinstance(pre_layer, Dense) and isinstance(next_layer, Dense):
        return dense_to_wider_layer(pre_layer, next_layer, n_add)
    elif isinstance(pre_layer, (Conv1D, Conv2D, Conv3D)) and isinstance(next_layer, (Conv1D, Conv2D, Conv3D)):
        return conv_to_wider_layer(pre_layer, next_layer, n_add)
    raise ValueError("Layer must be a Dense or Convolution Layer")



import os
import pickle

from keras.losses import categorical_crossentropy
from keras.models import load_model
from keras.optimizers import Adadelta

from autokeras import constant
from autokeras.generator import RandomConvClassifierGenerator, DefaultClassifierGenerator
from autokeras.utils import extract_config
from autokeras.net_transformer import transform
from autokeras.utils import ModelTrainer


class Searcher:
    def __init__(self, n_classes, input_shape, path, verbose):
        self.n_classes = n_classes
        self.input_shape = input_shape
        self.verbose = verbose
        self.history_configs = []
        self.history = []
        self.path = path
        self.model_count = 0

    def search(self, x_train, y_train, x_test, y_test):
        pass

    def load_best_model(self):
        model_id = max(self.history, key=lambda x: x['accuracy'])['model_id']
        return load_model(os.path.join(self.path, str(model_id) + '.h5'))

    def add_model(self, model, x_train, y_train, x_test, y_test):
        model.compile(loss=categorical_crossentropy,
                      optimizer=Adadelta(),
                      metrics=['accuracy'])
        if self.verbose:
            model.summary()
        ModelTrainer(model, x_train, y_train, x_test, y_test, self.verbose).train_model()
        loss, accuracy = model.evaluate(x_test, y_test, verbose=self.verbose)
        model.save(os.path.join(self.path, str(self.model_count) + '.h5'))
        self.history.append({'model_id': self.model_count, 'loss': loss, 'accuracy': accuracy})
        self.history_configs.append(extract_config(model))
        self.model_count += 1
        pickle.dump(self, open(os.path.join(self.path, 'searcher'), 'wb'))


class RandomSearcher(Searcher):
    def __init__(self, n_classes, input_shape, path, verbose):
        super().__init__(n_classes, input_shape, path, verbose)

    def search(self, x_train, y_train, x_test, y_test):
        # First model is randomly generated.
        while self.model_count < constant.MAX_MODEL_NUM:
            model = RandomConvClassifierGenerator(self.n_classes, self.input_shape).generate()
            self.add_model(model, x_train, y_train, x_test, y_test)

        return self.load_best_model()


class HillClimbingSearcher(Searcher):
    def __init__(self, n_classes, input_shape, path, verbose):
        super().__init__(n_classes, input_shape, path, verbose)

    def _remove_duplicate(self, models):
        """
        Remove the duplicate in the history_models
        :param models:
        :return:
        """
        ans = []
        for model_a in models:
            model_a_config = extract_config(model_a)
            if model_a_config not in self.history_configs:
                ans.append(model_a)
        return ans

    def search(self, x_train, y_train, x_test, y_test):
        # First model is randomly generated.
        if not self.history:
            model = DefaultClassifierGenerator(self.n_classes, self.input_shape).generate()
            self.add_model(model, x_train, y_train, x_test, y_test)

        optimal_accuracy = 0.0
        while self.model_count < constant.MAX_MODEL_NUM:
            model = self.load_best_model()
            new_models = self._remove_duplicate(transform(model))
            for model in new_models:
                self.history_configs.append(extract_config(model))

            for model in new_models:
                if self.model_count < constant.MAX_MODEL_NUM:
                    self.add_model(model, x_train, y_train, x_test, y_test)

            max_accuracy = max(self.history, key=lambda x: x['accuracy'])['accuracy']
            if max_accuracy <= optimal_accuracy:
                break
            optimal_accuracy = max_accuracy

        return self.load_best_model()



from keras.layers import Conv3D, Conv2D, Conv1D, Dense

CONV_FUNC_LIST = [Conv1D, Conv2D, Conv3D]
WEIGHTED_LAYER_FUNC_LIST = CONV_FUNC_LIST + [Dense]
MAX_MODEL_NUM = 100
MAX_ITER_NUM = 100000
MIN_LOSS_DEC = 1e-4
MAX_NO_IMPROVEMENT_NUM = 10
DEFAULT_SAVE_PATH = '/tmp/autokeras/'
EPOCHS_EACH = 10

LAYER_ATTR = {'Dense': ['units', 'activation'],
              'Dropout': ['rate'],
              'MaxPooling1D': ['pool_size'],
              'MaxPooling2D': ['pool_size'],
              'MaxPooling3D': ['pool_size'],
              'Conv1D': ['filters', 'activation', 'kernel_size'],
              'Conv2D': ['filters', 'activation', 'kernel_size'],
              'Conv3D': ['filters', 'activation', 'kernel_size'],
              'Flatten': []
              }



from random import randint, random

from keras.layers import MaxPooling1D, MaxPooling2D, MaxPooling3D, Dropout, Flatten, Dense
from keras.losses import categorical_crossentropy
from keras.models import Sequential
from keras.optimizers import Adadelta, Adam

from autokeras.utils import get_conv_layer_func


class ClassifierGenerator:
    def __init__(self, n_classes, input_shape):
        self.n_classes = n_classes
        self.input_shape = input_shape
        if len(self.input_shape) > 4:
            raise ValueError('The input dimension is too high.')
        if len(self.input_shape) < 2:
            raise ValueError('The input dimension is too low.')

    def _get_pool_layer_func(self):
        pool_funcs = [MaxPooling1D, MaxPooling2D, MaxPooling3D]
        return pool_funcs[len(self.input_shape) - 2]

    def _get_shape(self, dim_size):
        temp_list = [(dim_size,), (dim_size, dim_size), (dim_size, dim_size, dim_size)]
        return temp_list[len(self.input_shape) - 2]


class DefaultClassifierGenerator(ClassifierGenerator):
    def __init__(self, n_classes, input_shape):
        super().__init__(n_classes, input_shape)

    def generate(self):
        model = Sequential()
        pool = self._get_pool_layer_func()
        conv = get_conv_layer_func(len(self._get_shape(3)))
        model.add(conv(32, kernel_size=self._get_shape(3),
                       activation='relu',
                       padding='same',
                       input_shape=self.input_shape))
        model.add(conv(64, self._get_shape(3),
                       padding='same',
                       activation='relu'))
        model.add(pool(pool_size=self._get_shape(2)))
        model.add(Dropout(0.25))
        model.add(Flatten())
        model.add(Dense(128, activation='relu'))
        model.add(Dropout(0.5))
        model.add(Dense(self.n_classes, activation='softmax'))

        model.compile(loss=categorical_crossentropy,
                      optimizer=Adadelta(),
                      metrics=['accuracy'])
        return model


class RandomConvClassifierGenerator(ClassifierGenerator):
    def __init__(self, n_classes, input_shape):
        super().__init__(n_classes, input_shape)

    def generate(self):
        conv_num = randint(1, 10)
        dense_num = randint(1, 10)
        dropout_rate = random()
        filter_size = randint(1, 2) * 2 + 1
        pool_size = randint(2, 3)
        filter_shape = self._get_shape(filter_size)
        pool_shape = self._get_shape(pool_size)
        pool = self._get_pool_layer_func()
        conv = get_conv_layer_func(len(filter_shape))

        model = Sequential()
        for i in range(conv_num):
            kernel_num = randint(10, 30)
            if i == 0:
                model.add(conv(kernel_num,
                               input_shape=self.input_shape,
                               kernel_size=filter_shape,
                               activation='relu',
                               padding='same'))
            else:
                model.add(conv(kernel_num,
                               kernel_size=filter_shape,
                               activation='relu',
                               padding='same'))
            if random() > 0.5:
                model.add(pool(pool_size=pool_shape, padding='same'))
            if random() > 0.5:
                model.add(Dropout(dropout_rate))
        model.add(Flatten())
        for i in range(dense_num):
            node_num = randint(128, 1024)
            model.add(Dense(node_num, activation='relu'))
            if random() > 0.5:
                model.add(Dropout(dropout_rate))
        model.add(Dense(self.n_classes, activation='softmax'))
        model.compile(loss='categorical_crossentropy',
                      optimizer=Adam(),
                      metrics=['accuracy'])
        return model



from distutils.core import setup

setup(
    name='autokeras',
    packages=['autokeras'],  # this must be the same as the name above
    install_requires=['pytest', 'numpy', 'keras', 'scikit-learn', 'tensorflow'],
    version='0.0.2',
    description='Automated Machine Learning with Keras',
    author='Haifeng Jin',
    author_email='jhfjhfj1@gmail.com',
    url='https://github.com/jhfjhfj1/auto-keras',
    download_url='https://github.com/jhfjhfj1/autokeras/archive/0.0.2.tar.gz',
    keywords=['automl'],  # arbitrary keywords
    classifiers=[]
)



import os
from keras.layers import Conv1D, Conv2D, Conv3D
from autokeras.constant import CONV_FUNC_LIST, LAYER_ATTR
from autokeras import constant
from keras import backend


def is_conv_layer(layer):
    return isinstance(layer, tuple(CONV_FUNC_LIST))


def get_conv_layer_func(n_dim):
    conv_layer_functions = [Conv1D, Conv2D, Conv3D]
    if n_dim > 3:
        raise ValueError('The input dimension is too high.')
    if n_dim < 1:
        raise ValueError('The input dimension is too low.')
    return conv_layer_functions[n_dim - 1]


class ModelTrainer:
    def __init__(self, model, x_train, y_train, x_test, y_test, verbose):
        self.model = model
        self.x_train = x_train
        self.y_train = y_train
        self.x_test = x_test
        self.y_test = y_test
        self.verbose = verbose
        self.training_losses = []
        self.minimum_loss = None
        self._no_improvement_count = 0

    def _converged(self, loss):
        self.training_losses.append(loss)
        if loss > (self.minimum_loss - constant.MIN_LOSS_DEC):
            self._no_improvement_count += 1
        else:
            self._no_improvement_count = 0

        if loss < self.minimum_loss:
            self.minimum_loss = loss

        return self._no_improvement_count > constant.MAX_NO_IMPROVEMENT_NUM

    def train_model(self):
        self.training_losses = []
        self._no_improvement_count = 0
        self.minimum_loss = float('inf')
        for _ in range(constant.MAX_ITER_NUM):
            self.model.fit(self.x_train, self.y_train,
                           batch_size=min(self.x_train.shape[0], 200),
                           epochs=constant.EPOCHS_EACH,
                           verbose=self.verbose)
            loss, _ = self.model.evaluate(self.x_test, self.y_test, verbose=self.verbose)
            if self._converged(loss):
                break
        return self.minimum_loss


def copy_layer(layer):
    new_layer = layer.__class__.from_config(layer.get_config())
    if new_layer is None:
        raise ValueError("There must be a Dense or Convolution Layer")
    return new_layer


def extract_config(network):
    config = {'type': [], 'config': []}
    for layer in network.layers:
        name = type(layer).__name__
        config['type'].append(name)
        layer_config = layer.get_config()
        important_attr = {}
        for attr in LAYER_ATTR[name]:
            important_attr[attr] = layer_config[attr]
        config['config'].append(important_attr)
    return config


def ensure_dir(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


def ensure_file_dir(path):
    ensure_dir(os.path.dirname(path))


def reset_weights(model):
    session = backend.get_session()
    for layer in model.layers:
        if hasattr(layer, 'kernel_initializer'):
            layer.kernel.initializer.run(session=session)



from autokeras.constant import WEIGHTED_LAYER_FUNC_LIST
from autokeras.layer_transformer import to_deeper_layer, to_wider_layer

from keras.models import Sequential
from keras.layers import Dense, Conv1D, Conv2D, Conv3D

from autokeras.utils import copy_layer


def get_next_dense_conv(start, layers):
    """Start to search next dense or conv.

    long description.

    :param start:
    :param layers:
    :return:
    """
    new_next_wider_layer = None
    ind = None
    for j in range(start + 1, len(layers)):
        if isinstance(layers[j], (
                Dense, Conv1D, Conv2D, Conv3D)):
            new_next_wider_layer = layers[j]
            ind = j
            break
    if new_next_wider_layer is None:
        raise ValueError("There must be a Corresponding Dense or Convolution Layer")
    return new_next_wider_layer, ind


def replace_layers(model, level_list, layer_list):
    new_model = Sequential()
    new_layer_list = []

    for layer in model.layers:
        new_layer_list.append(copy_layer(layer))

    for index, level in enumerate(level_list):
        new_layer_list[level] = layer_list[index]

    for index, layer in enumerate(new_layer_list):
        new_model.add(layer)

    for index, layer in enumerate(new_model.layers):
        if index not in level_list:
            layer.set_weights(model.layers[index].get_weights())

    return new_model


def insert_layer(model, level, new_layer):
    new_model = Sequential()

    for index, layer in enumerate(model.layers):
        new_model.add(copy_layer(layer))
        new_model.layers[-1].set_weights(layer.get_weights())
        if index == level:
            new_model.add(new_layer)

    return new_model


def to_deeper_model(model, level):
    new_deeper_layer = to_deeper_layer(model.layers[level])
    return insert_layer(model, level, new_deeper_layer)


def to_wider_model(model, level):
    next_wider_layer, ind = get_next_dense_conv(level, model.layers)
    new_wider_layer, new_next_wider_layer = to_wider_layer(model.layers[level], next_wider_layer, 1)
    return replace_layers(model, [level, ind], [new_wider_layer, new_next_wider_layer])


def transform(model):
    print('Function called.')
    models = []
    layers = model.layers
    for index in range(len(layers) - 1):
        if isinstance(layers[index], tuple(WEIGHTED_LAYER_FUNC_LIST)):
            models.append(to_deeper_model(model, index))
            models.append(to_wider_model(model, index))
    return models
